import os

import llvmlite.binding as llvm
import llvmlite.ir

from lleaves.compiler.ast import parse_to_ast
from lleaves.compiler.codegen import gen_forest


def compile_to_module(file_path, fblocksize=34, finline=True, raw_score=False):
    forest = parse_to_ast(file_path)
    forest.raw_score = raw_score

    ir = llvmlite.ir.Module(name="forest")
    gen_forest(forest, ir, fblocksize)

    ir.triple = llvm.get_process_triple()
    module = llvm.parse_assembly(str(ir))
    module.name = str(file_path)
    module.verify()

    if os.environ.get("LLEAVES_PRINT_UNOPTIMIZED_IR") == "1":
        print(module)

    # Create optimizer
    pmb = llvm.PassManagerBuilder()
    pmb.opt_level = 3

    if finline:
        # if inline_threshold is set LLVM inlines, precise value doesn't seem to matter
        pmb.inlining_threshold = 1

    pm_module = llvm.ModulePassManager()
    # Add optimization passes to module-level optimizer
    pmb.populate(pm_module)

    # single pass only, compiler optimizations don't bring much speedup and take time
    pm_module.run(module)

    if os.environ.get("LLEAVES_PRINT_OPTIMIZED_IR") == "1":
        print(module)

    return module



import json
import os
from ctypes import POINTER, c_double
from typing import List, Optional

import numpy as np

try:
    from pandas import DataFrame as pd_DataFrame
except ImportError:

    class pd_DataFrame:
        """Dummy class for pandas.DataFrame."""

        pass


def _dataframe_to_ndarray(data, pd_traintime_categories: List[List]):
    """
    Converts the given dataframe into a 2D numpy array and converts categorical columns to float.

    Categoricals present in the dataframe are mapped to their float IDs. The `pd_traintime_categories` are used
    to ensure this categorical -> ID mapping is the same as it was in the training dataset.

    :param data: 2D pandas dataframe.
    :param pd_traintime_categories: For each column a list of all categories in this column.
        The ordering of columns and of categories within each column should match the training dataset.

        Example (two columns with two categories each): ``[["a", "b"], ["b", "a"]]``.
        These columns are different and will result in two different mappings ("a" -> 0.0 vs "a" -> 1.0).
    :return: 2D np.ndarray, dtype float64 or float32
    """
    cat_cols = list(data.select_dtypes(include=["category"]).columns)
    if len(cat_cols) != len(pd_traintime_categories):
        raise ValueError(
            "The categorical columns in the dataset don't match the categorical columns during training!"
            f"Train had {len(pd_traintime_categories)} categorical columns, data has {len(cat_cols)}"
        )
    if len(cat_cols):
        data = data.copy()
        for col, category in zip(cat_cols, pd_traintime_categories):
            # we use set_categories to get the same (category -> code) mapping that we used during train
            if list(data[col].cat.categories) != list(category):
                data[col] = data[col].cat.set_categories(category)
        # apply (category -> code) mapping. Categories become floats
        data[cat_cols] = (
            data[cat_cols].apply(lambda x: x.cat.codes).replace({-1: np.nan})
        )
    data = data.values
    if data.dtype != np.float64 and data.dtype != np.float32:
        data = data.astype(np.float64)
    return data


def data_to_ndarray(data, pd_traintime_categories: Optional[List[List]] = None):
    """
    Convert the given data to a numpy ndarray

    For pandas dataframes categories are mapped to floats.
    This mapping needs to be the same as it was during model training, which is achieved via ``pandas_categorical``.

    Example for two columns with two categories each: ``pd_traintime_categories = [["a", "b"], ["b", "a"]]``.
    These are two different columns and result in different mappings: "a" -> 0.0, "b" -> 1.0, vs "b" -> 0.0, "a" -> 1.0.

    LightGBM generates this list of lists at traintime like so::

      pd_traintime_categories = [
        list(df[col].cat.categories)
        for col in df.select_dtypes(include=['category']).columns
      ]

    The result is appended via ``json.dump`` to the model.txt under the 'pandas_categorical' key.
    You can extract it from there using :func:`lleaves.data_processing.extract_pandas_traintime_categories`.

    :param data: Pandas dataframe, numpy array or Python list. No dimension checking occurs. If a dataframe is passed
        the number of categorical columns needs to equal ``len(pd_traintime_categories)``.
    :param pd_traintime_categories: For each categorical column in dataframe, a list of its categories.
        The ordering of columns and of categories within each column should match the training dataset.
        Ignored if data is not a pandas DataFrame.

    :return: numpy ndarray
    """
    if isinstance(data, np.ndarray):
        data = data
    elif isinstance(data, pd_DataFrame):
        data = _dataframe_to_ndarray(data, pd_traintime_categories)
    elif isinstance(data, list):
        data = np.array(data, dtype=np.float64)
    else:
        raise ValueError(
            f"Expecting numpy.ndarray, pandas.DataFrame or Python list, got {type(data)}"
        )

    return data


def ndarray_to_ptr(data):
    """
    Takes a 2D numpy array, converts to float64 if necessary and returns a pointer

    :param data: 2D numpy array. Copying is avoided if possible.
    :return: pointer to 1D array of dtype float64.
    """
    # ravel makes sure we get a contiguous array in memory and not some strided View
    data = data.astype(np.float64, copy=False, casting="same_kind").ravel()
    ptr = data.ctypes.data_as(POINTER(c_double))
    return ptr


def extract_pandas_traintime_categories(file_path):
    """
    Scan the model.txt from the back to extract the 'pandas_categorical' field.

    This is a list of lists that stores the ordering of categories from the pd.DataFrame used for training.
    Storing this list is necessary as LightGBM encodes categories as integer indices and we need to guarantee that
    the mapping (<category string> -> <integer idx>) is the same during inference as it was during training.

    - Example (pandas categoricals were present in training):
      ``pandas_categorical:[["a", "b", "c"], ["b", "c", "d"], ["w", "x", "y", "z"]]``
    - Example (no pandas categoricals during training):
      ``pandas_categorical:[]`` or ``pandas_categorical=null``

    :param file_path: path to model.txt
    :return: list of list. For each pd.categorical column encountered during training, a list of the categories.
    """
    pandas_key = "pandas_categorical:"
    max_offset = os.path.getsize(file_path)
    stepsize = min(1024, max_offset - 1)
    current_offset = stepsize
    lines = []
    # seek backwards from end of file until we have two lines
    # the (pen)ultimate line should be pandas_categorical:XXX
    with open(file_path, "rb") as f:
        while len(lines) < 2 and current_offset < max_offset:
            if current_offset > max_offset:
                current_offset = max_offset
            # read <current_offset>-many Bytes from end of file
            f.seek(-current_offset, os.SEEK_END)
            lines = f.readlines()
            current_offset *= 2

    # pandas_categorical has to be present in the ultimate or penultimate line. Else the model.txt is malformed.
    if len(lines) >= 2:
        last_line = lines[-1].decode().strip()
        if not last_line.startswith(pandas_key):
            last_line = lines[-2].decode().strip()
        if last_line.startswith(pandas_key):
            return json.loads(last_line[len(pandas_key) :])
    raise ValueError("Ill formatted model file!")


def extract_model_global_features(file_path):
    """
    Extract number of features, number of classes and number of trees of this model

    :param file_path: path to model.txt
    :return: dict with "n_args", "n_classes", "n_trees"
    """
    res = {}
    with open(file_path, "r") as f:
        for _ in range(3):
            line = f.readline()
            while line and not line.startswith(
                ("max_feature_idx", "num_class", "tree_sizes")
            ):
                line = f.readline()

            if line.startswith("max_feature_idx"):
                res["n_feature"] = int(line.split("=")[1]) + 1
            elif line.startswith("num_class"):
                res["n_class"] = int(line.split("=")[1])
            elif line.startswith("tree_sizes"):
                # `tree_sizes=123 123 123 123`
                res["n_trees"] = len(line.split("=")[1].split(" "))
            else:
                raise ValueError("Ill formatted model file!")
    return res



from dataclasses import dataclass

from llvmlite import ir

from lleaves.compiler.utils import ISSUE_ERROR_MSG, MissingType

BOOL = ir.IntType(bits=1)
DOUBLE = ir.DoubleType()
FLOAT = ir.FloatType()
INT_CAT = ir.IntType(bits=32)
INT = ir.IntType(bits=32)
ZERO_V = ir.Constant(BOOL, 0)
FLOAT_POINTER = ir.PointerType(FLOAT)
DOUBLE_PTR = ir.PointerType(DOUBLE)


def iconst(value):
    return ir.Constant(INT, value)


def fconst(value):
    return ir.Constant(FLOAT, value)


def dconst(value):
    return ir.Constant(DOUBLE, value)


@dataclass
class LTree:
    """Class for the LLVM function of a tree paired with relevant non-LLVM context"""

    llvm_function: ir.Function
    class_id: int


def gen_forest(forest, module, fblocksize):
    """
    Populate the passed IR module with code for the forest.

    Overview of generated IR
    ---------------------------
    The forest is represented by the @forest_root function, which is called from Python.
    This function loops over every row of the input array. For each row:
    - Load all attributes, cast categorical attributes to INT.
    - Iteratively call each @tree_<index> function. This function returns a DOUBLE. The results
      of all @tree_<index> calls are summed up into a result variable.
    - The final result variable is run through the objective function (eg sigmoid)
      and stored in the results array passed by the caller.

    The actual IR is slightly more complicated, because lleaves implements instruction cache blocking.
    The full set of tree_functions is divided into chunks, with each chunk containing a subset of tree_functions.
    For each chunk we process every row of the input array in sequence which minimizes icache misses.
    Pseudo-code for tree with 100 tree_functions and chunks of size 50:
    for tree in tree_funcs[0:50]:
        for row in range(len(input)):
           result[row] += tree(input[row])
    for tree in tree_funcs[50:100]:
        for row in range(len(input)):
           result[row] += tree(input[row])
    ...

    For each tree in the forest there is a @tree_<index> function which takes all attributes as arguments

    For each node there are 0-2 blocks in the @tree_<index> function.
    - Decision node (categorical): 2 Blocks, 1 for the node, 1 for the categorical bitset-comparison
    - Decision node (numerical): 1 Block.
    - Leaf node: 0-1 Blocks. If a decision node has only leaves as children we fuse both leaves into
      a single switch instr in the decision node's block.
    Each node cbranches to the child node's block.

    :return: None
    """

    # entry function called from Python
    root_func = ir.Function(
        module,
        ir.FunctionType(ir.VoidType(), (DOUBLE_PTR, DOUBLE_PTR, INT, INT)),
        name="forest_root",
    )

    def make_tree(tree):
        # declare the function for this tree
        func_dtypes = (INT_CAT if f.is_categorical else DOUBLE for f in tree.features)
        scalar_func_t = ir.FunctionType(DOUBLE, func_dtypes)
        tree_func = ir.Function(module, scalar_func_t, name=str(tree))
        tree_func.linkage = "private"
        # populate function with IR
        gen_tree(tree, tree_func)
        return LTree(llvm_function=tree_func, class_id=tree.class_id)

    tree_funcs = [make_tree(tree) for tree in forest.trees]

    if forest.n_classes > 1:
        # better locality by running trees for each class together
        tree_funcs.sort(key=lambda t: t.class_id)

    _populate_forest_func(forest, root_func, tree_funcs, fblocksize)


def gen_tree(tree, tree_func):
    """generate code for tree given the function, recursing into nodes"""
    node_block = tree_func.append_basic_block(name=str(tree.root_node))
    gen_node(tree_func, node_block, tree.root_node)


def gen_node(func, node_block, node):
    """generate code for node, recursing into children"""
    if node.is_leaf:
        _gen_leaf_node(node_block, node)
    else:
        _gen_decision_node(func, node_block, node)


def _gen_leaf_node(node_block, leaf):
    """populate block with leaf's return value"""
    builder = ir.IRBuilder(node_block)
    builder.ret(dconst(leaf.value))


def _gen_decision_node(func, node_block, node):
    """generate code for decision node, recursing into children"""
    builder = ir.IRBuilder(node_block)

    # optimization for node where both children are leaves (switch instead of cbranch)
    is_fused_double_leaf_node = node.left.is_leaf and node.right.is_leaf
    if is_fused_double_leaf_node:
        left_block = None
        right_block = None
        # categorical nodes have a fastpath which can branch-right early
        # so they still need a right block
        if node.decision_type.is_categorical:
            right_block = func.append_basic_block(name=str(node.right))
    else:
        left_block = func.append_basic_block(name=str(node.left))
        right_block = func.append_basic_block(name=str(node.right))

    # populate this node's block up to the terminal statement
    if node.decision_type.is_categorical:
        bitset_comp_block = builder.append_basic_block(str(node) + "_cat_bitset_comp")
        bitset_builder = ir.IRBuilder(bitset_comp_block)
        comp = _populate_categorical_node_block(
            func, builder, bitset_builder, node, bitset_comp_block, right_block
        )
        builder = bitset_builder
    else:
        comp = _populate_numerical_node_block(func, builder, node)

    # finalize this node's block with a terminal statement
    if is_fused_double_leaf_node:
        ret = builder.select(comp, dconst(node.left.value), dconst(node.right.value))
        builder.ret(ret)
    else:
        builder.cbranch(comp, left_block, right_block)

    # populate generated child blocks
    if left_block:
        gen_node(func, left_block, node.left)
    if right_block:
        gen_node(func, right_block, node.right)


def _populate_instruction_block(
    forest, root_func, tree_funcs, setup_block, next_block, eval_obj_func
):
    """Generates an instruction_block: loops over all input data and evaluates its chunk of tree_funcs."""
    data_arr, out_arr, start_index, end_index = root_func.args

    # -- SETUP BLOCK
    builder = ir.IRBuilder(setup_block)
    loop_iter = builder.alloca(INT, 1, "loop-idx")
    builder.store(start_index, loop_iter)
    condition_block = root_func.append_basic_block("loop-condition")
    builder.branch(condition_block)
    # -- END SETUP BLOCK

    # -- CONDITION BLOCK
    builder = ir.IRBuilder(condition_block)
    comp = builder.icmp_signed("<", builder.load(loop_iter), end_index)
    core_block = root_func.append_basic_block("loop-core")
    builder.cbranch(comp, core_block, next_block)
    # -- END CONDITION BLOCK

    # -- CORE LOOP BLOCK
    builder = ir.IRBuilder(core_block)
    # build args arr, convert categoricals vars from float to int
    args = []
    loop_iter_reg = builder.load(loop_iter)

    n_args = ir.Constant(INT, forest.n_args)
    iter_mul_nargs = builder.mul(loop_iter_reg, n_args)
    idx = (builder.add(iter_mul_nargs, iconst(i)) for i in range(forest.n_args))
    raw_ptrs = [builder.gep(root_func.args[0], (c,)) for c in idx]
    # cast the categorical inputs to integer
    for feature, ptr in zip(forest.features, raw_ptrs):
        el = builder.load(ptr)
        if feature.is_categorical:
            args.append(builder.fptosi(el, INT_CAT))
        else:
            args.append(el)
    # iterate over each tree, sum up results
    results = [dconst(0.0) for _ in range(forest.n_classes)]
    for func in tree_funcs:
        tree_res = builder.call(func.llvm_function, args)
        results[func.class_id] = builder.fadd(tree_res, results[func.class_id])
    res_idx = builder.mul(iconst(forest.n_classes), loop_iter_reg)
    results_ptr = [
        builder.gep(out_arr, (builder.add(res_idx, iconst(class_idx)),))
        for class_idx in range(forest.n_classes)
    ]

    results = [
        builder.fadd(result, builder.load(result_ptr))
        for result, result_ptr in zip(results, results_ptr)
    ]
    if eval_obj_func:
        results = _populate_objective_func_block(
            builder,
            results,
            forest.objective_func,
            forest.objective_func_config,
            forest.raw_score,
        )
    for result, result_ptr in zip(results, results_ptr):
        builder.store(result, result_ptr)

    tmpp1 = builder.add(loop_iter_reg, iconst(1))
    builder.store(tmpp1, loop_iter)
    builder.branch(condition_block)
    # -- END CORE LOOP BLOCK


def _populate_forest_func(forest, root_func, tree_funcs, fblocksize):
    """Populate root function IR for forest"""

    assert fblocksize > 0
    # generate the setup-blocks upfront, so each instruction_block can be passed its successor
    instr_blocks = [
        (
            root_func.append_basic_block("instr-block-setup"),
            tree_funcs[i : i + fblocksize],
        )
        for i in range(0, len(tree_funcs), fblocksize)
    ]
    term_block = root_func.append_basic_block("term")
    ir.IRBuilder(term_block).ret_void()
    for i, (setup_block, tree_func_chunk) in enumerate(instr_blocks):
        next_block = instr_blocks[i + 1][0] if i < len(instr_blocks) - 1 else term_block
        eval_objective_func = next_block == term_block
        _populate_instruction_block(
            forest,
            root_func,
            tree_func_chunk,
            setup_block,
            next_block,
            eval_objective_func,
        )


def _populate_objective_func_block(
    builder, args, objective: str, objective_config: str, raw_score: bool
):
    """
    Takes the objective function specification and generates the code for it into the builder
    """
    llvm_exp = builder.module.declare_intrinsic("llvm.exp", (DOUBLE,))
    llvm_log = builder.module.declare_intrinsic("llvm.log", (DOUBLE,))
    llvm_copysign = builder.module.declare_intrinsic(
        "llvm.copysign", (DOUBLE, DOUBLE), ir.FunctionType(DOUBLE, (DOUBLE, DOUBLE))
    )

    def _populate_sigmoid(alpha):
        if alpha <= 0:
            raise ValueError(f"Sigmoid parameter needs to be >0, is {alpha}")

        # 1 / (1 + exp(- alpha * x))
        inner = builder.fmul(dconst(-alpha), args[0])
        exp = builder.call(llvm_exp, [inner])
        denom = builder.fadd(dconst(1.0), exp)
        return builder.fdiv(dconst(1.0), denom)

    # raw score means we don't need to add the objective function
    if raw_score:
        return args

    if objective == "binary":
        alpha = objective_config.split(":")[1]
        result = _populate_sigmoid(float(alpha))
    elif objective in ("xentropy", "cross_entropy"):
        result = _populate_sigmoid(1.0)
    elif objective in ("xentlambda", "cross_entropy_lambda"):
        # naive implementation which will be numerically unstable for small x.
        # should be changed to log1p
        exp = builder.call(llvm_exp, [args[0]])
        result = builder.call(llvm_log, [builder.fadd(dconst(1.0), exp)])
    elif objective in ("poisson", "gamma", "tweedie"):
        result = builder.call(llvm_exp, [args[0]])
    elif objective in (
        "regression",
        "regression_l1",
        "huber",
        "fair",
        "quantile",
        "mape",
    ):
        if objective_config and "sqrt" in objective_config:
            arg = args[0]
            result = builder.call(llvm_copysign, [builder.fmul(arg, arg), arg])
        else:
            result = args[0]
    elif objective in ("lambdarank", "rank_xendcg", "custom"):
        result = args[0]
    elif objective == "multiclass":
        assert len(args)
        # TODO Might profit from vectorization, needs testing
        result = [builder.call(llvm_exp, [arg]) for arg in args]

        denominator = dconst(0.0)
        for r in result:
            denominator = builder.fadd(r, denominator)

        result = [builder.fdiv(r, denominator) for r in result]
    else:
        raise ValueError(
            f"Objective '{objective}' not yet implemented. {ISSUE_ERROR_MSG}"
        )
    return result if len(args) > 1 else [result]


def _populate_categorical_node_block(
    func, builder, bitset_comp_builder, node, bitset_comp_block, right_block
):
    """Populate block with IR for categorical node"""
    val = func.args[node.split_feature]

    # For categoricals, processing NaNs happens through casting them via fptosi in the Forest root
    # NaNs become negative max_val, which never exists in the Bitset, so they always go right

    # Find in bitset
    # First, check value > max categorical
    comp = builder.icmp_unsigned(
        "<",
        val,
        iconst(32 * len(node.cat_threshold)),
    )
    builder.cbranch(comp, bitset_comp_block, right_block)

    idx = bitset_comp_builder.udiv(val, iconst(32))
    bit_vecs = ir.Constant(
        ir.VectorType(INT, len(node.cat_threshold)),
        [ir.Constant(INT, i) for i in node.cat_threshold],
    )
    shift = bitset_comp_builder.urem(val, iconst(32))
    # pick relevant bitvector
    bit_vec = bitset_comp_builder.extract_element(bit_vecs, idx)
    # check bitvector contains
    bit_entry = bitset_comp_builder.lshr(bit_vec, shift)
    comp = bitset_comp_builder.trunc(bit_entry, BOOL)
    return comp


def _populate_numerical_node_block(func, builder, node):
    """populate block with IR for numerical node"""
    val = func.args[node.split_feature]

    thresh = ir.Constant(DOUBLE, node.threshold)
    missing_t = node.decision_type.missing_type

    # If missingType != MNaN, LightGBM treats NaNs values as if they were 0.0.
    # So for MZero, NaNs get treated like missing values.
    # But for MNone, NaNs get treated as the literal value 0.0.
    # default_left decides where to go when a missing value is encountered
    # for MNone handle NaNs by adjusting default_left to make sure NaNs go where 0.0 would have gone.
    # for MZero we handle NaNs in the IR
    if node.decision_type.missing_type == MissingType.MNone:
        default_left = 0.0 <= node.threshold
    else:
        default_left = node.decision_type.is_default_left

    # MissingType.MZero: Treat 0s (and NaNs) as missing values
    if default_left:
        if missing_t != MissingType.MZero or (
            missing_t == MissingType.MZero and 0.0 <= node.threshold
        ):
            # unordered cmp: we'll get True (and go left) if any arg is qNaN
            comp = builder.fcmp_unordered("<=", val, thresh)
        else:
            is_missing = builder.fcmp_unordered("==", val, fconst(0.0))
            less_eq = builder.fcmp_unordered("<=", val, thresh)
            comp = builder.or_(is_missing, less_eq)
    else:
        if missing_t != MissingType.MZero or (
            missing_t == MissingType.MZero and node.threshold < 0.0
        ):
            # ordered cmp: we'll get False (and go right) if any arg is qNaN
            comp = builder.fcmp_ordered("<=", val, thresh)
        else:
            is_missing = builder.fcmp_unordered("==", val, fconst(0.0))
            greater = builder.fcmp_ordered(">", val, thresh)
            comp = builder.not_(builder.or_(is_missing, greater))
    return comp



import itertools

from lleaves.compiler.ast.nodes import DecisionNode, Forest, LeafNode, Tree
from lleaves.compiler.ast.scanner import scan_model_file
from lleaves.compiler.utils import DecisionType

"""
The parser takes the results from the scanner and transforms them into an
Abstract-Syntax Tree (AST).
It builds up the Graph for the DecisionTree-Forest, consisting of decision-nodes and leaf-nodes.
"""


class Feature:
    """
    Represents one feature (= column) that is passed to the tree function or forest function.
    """

    def __init__(self, is_categorical):
        self.is_categorical = is_categorical


def _parse_tree_to_ast(tree_struct, features, class_id):
    n_nodes = len(tree_struct["decision_type"])
    leaves = [
        LeafNode(idx, value) for idx, value in enumerate(tree_struct["leaf_value"])
    ]

    # Create the nodes using all non-specific data
    # categorical nodes are finalized later
    nodes = [
        DecisionNode(
            idx,
            split_feature,
            threshold,
            DecisionType(decision_type_id),
            left_idx,
            right_idx,
        )
        for idx, (
            split_feature,
            threshold,
            decision_type_id,
            left_idx,
            right_idx,
        ) in enumerate(
            zip(
                tree_struct["split_feature"],
                tree_struct["threshold"],
                tree_struct["decision_type"],
                tree_struct["left_child"],
                tree_struct["right_child"],
            )
        )
    ]
    assert len(nodes) == n_nodes

    categorical_nodes = [
        idx
        for idx, decision_type_id in enumerate(tree_struct["decision_type"])
        if DecisionType(decision_type_id).is_categorical
    ]

    for idx in categorical_nodes:
        node = nodes[idx]
        thresh = int(node.threshold)
        # pass just the relevant vector entries
        start = tree_struct["cat_boundaries"][thresh]
        end = tree_struct["cat_boundaries"][thresh + 1]
        node.finalize_categorical(
            cat_threshold=tree_struct["cat_threshold"][start:end],
        )

    for node in nodes:
        # in the model_file.txt, the outgoing left + right nodes are specified
        # via their index in the list. negative numbers are leaves, positive numbers
        # are other nodes
        children = [
            leaves[abs(idx) - 1] if idx < 0 else nodes[idx]
            for idx in (node.left_idx, node.right_idx)
        ]
        node.add_children(*children)

    for node in nodes:
        node.validate()

    if nodes:
        return Tree(tree_struct["Tree"], nodes[0], features, class_id)
    else:
        # special case for when tree is just single leaf
        assert len(leaves) == 1
        return Tree(tree_struct["Tree"], leaves[0], features, class_id)


def parse_to_ast(model_path):
    scanned_model = scan_model_file(model_path)

    n_args = scanned_model["general_info"]["max_feature_idx"] + 1
    n_classes = scanned_model["general_info"]["num_class"]
    assert n_classes == scanned_model["general_info"]["num_tree_per_iteration"]
    objective = scanned_model["general_info"]["objective"]
    objective_func = objective[0]
    objective_func_config = objective[1] if len(objective) > 1 else None
    features = [
        Feature(is_categorical_feature(x))
        for x in scanned_model["general_info"]["feature_infos"]
    ]
    assert n_args == len(features), "Ill formed model file"

    trees = [
        _parse_tree_to_ast(scanned_tree, features, class_id)
        for scanned_tree, class_id in zip(
            scanned_model["trees"], itertools.cycle(range(n_classes))
        )
    ]
    assert len(trees) % n_classes == 0, "Ill formed model file"
    return Forest(trees, features, n_classes, objective_func, objective_func_config)


def is_categorical_feature(feature_info: str):
    """
    :param feature_info: one entry from the model.txt 'feature_infos' field
    """
    # Feature infos for floats look like [x.xxxx:y.yyyy]
    # for categoricals like X:Y:Z:
    return not feature_info.startswith("[")
